//
// adi_ttcp_transfer.c
//
// Contains the code to execute TRANSFER requests.
//
// The entrypoint is TTCP_DispatchTransfer() which sets up the initial
// conditions for a transfer and then calls the appropriate function to
// execute it.
//
#include "config.h"
#include "adi_ttcp.h"


// an expression that returns non-zero if a transfer has no maximum period or if
// the period has not expired, and 0 if the period has expired
#define notExpired(p)                                                        \
    (((p)->m_Cmd.m_period == 0)?                                             \
	1 :                                                                  \
	(TTCP_GetMillisecCount() - (p)->m_tStart) <= (p)->m_Cmd.m_period)


static int one = 1;         // for 4.3 BSD style setsockopt()
static int zero = 0;        // for 4.3 BSD style setsockopt()


/*
 *          N R E A D
 */
static int TTCP_Nread( PTEST_BLOCK pTBlk, int count )
{
   SOCKADDR_IN from;
   int len = sizeof(from);
   register int cnt;

   if( pTBlk->m_Cmd.m_Protocol == IPPROTO_UDP )
   {
      cnt = recvfrom( pTBlk->m_Socket_fd,
               pTBlk->m_pBuf, count, 0,
               (PSOCKADDR )&from,
               &len
            );

      pTBlk->m_Cmd.m_RemoteAddr.sin_addr.s_addr = from.sin_addr.s_addr;
      pTBlk->m_Cmd.m_RemoteAddr.sin_port = from.sin_port;

      pTBlk->m_numCalls++;
   }
   else
   {
      cnt = recv( pTBlk->m_Socket_fd, pTBlk->m_pBuf, count, 0 );

      if( cnt == SOCKET_ERROR )
      {
         int nError = TTCP_GetLastError();
      }

      pTBlk->m_numCalls++;
   }

   if( pTBlk->m_Cmd.m_bTouchRecvData && cnt > 0 )
   {
      register int c = cnt, sum;
      volatile char *b = pTBlk->m_pBuf;

      sum = 0;
      while (c--)
         sum += *b++;
   }

   return (cnt);
}

/*
 *          N W R I T E
 */
static int TTCP_Nwrite( PTEST_BLOCK pTBlk, int count )
{
   register int cnt;

   //
   // Introduce Write Delay After First Write Call
   //
   if( pTBlk->m_Cmd.m_nOptWriteDelay && pTBlk->m_numCalls > 0 )
   {
      TTCP_DelayMicrosec(pTBlk->m_Cmd.m_nOptWriteDelay);
   }

   if( pTBlk->m_Cmd.m_Protocol == IPPROTO_UDP )
   {
again:
      cnt = sendto(
		pTBlk->m_Socket_fd,
		pTBlk->m_pBuf,
		count,
		0,
		(PSOCKADDR )&pTBlk->m_Cmd.m_RemoteAddr,
		sizeof(pTBlk->m_Cmd.m_RemoteAddr)
	    );

      pTBlk->m_numCalls++;

      if( cnt == SOCKET_ERROR && TTCP_GetLastError() == TTCP_NO_BUFFERS )
      {
         TTCP_SleepMillisec(18);
         errno = 0;
         goto again;
      }
   }
   else
   {
      cnt = send( pTBlk->m_Socket_fd, pTBlk->m_pBuf, count, 0 );
      pTBlk->m_numCalls++;
   }

   return(cnt);
}

static char *outfmt( double b, char fmt, char* obuf )
{
   switch (fmt) {
    case 'G':
        sprintf(obuf, "%.2f GB", b / 1000.0 / 1000.0 / 1000.0);
        break;
    default:
    case 'K':
        sprintf(obuf, "%.2f KB", b / 1000.0);
        break;
    case 'M':
        sprintf(obuf, "%.2f MB", b / 1000.0 / 1000.0);
        break;
    case 'g':
        sprintf(obuf, "%.2f Gbit", b * 8.0 / 1000.0 / 1000.0 / 1000.0);
        break;
    case 'k':
        sprintf(obuf, "%.2f Kbit", b * 8.0 / 1000.0);
        break;
    case 'm':
        sprintf(obuf, "%.2f Mbit", b * 8.0 / 1000.0 / 1000.0);
        break;
    }

    return obuf;
}

static void TTCP_InitStatistics( PTEST_BLOCK pTBlk )
{
   pTBlk->m_numCalls = 0;    // # of I/O system calls
   pTBlk->m_nbytes = 0;      // bytes on net

   TTCP_TransferStarting( pTBlk );
   pTBlk->m_tStart = TTCP_GetMillisecCount();
}


static void TTCP_LogStatistics( PTEST_BLOCK pTBlk )
{
   PMSG_BLOCK pmsg = &pTBlk->m_Cmd.m_msgBlk;

   double realt;                 // user, real time (seconds)
   double idlet;                 // idle time during test (seconds)
   char buf[50];

   pTBlk->m_tFinish = TTCP_GetMillisecCount();
   TTCP_TransferStopping( pTBlk );

   realt = ((double )pTBlk->m_tFinish - (double )pTBlk->m_tStart)/1000;

   if( pTBlk->m_Cmd.m_Protocol == IPPROTO_UDP )
   {
      if( pTBlk->m_Cmd.m_bTransmit )
      {
	TTCP_AddMsg(
	    pmsg,
	    "  Statistics  : UDP -> %s:%d\r\n",
            INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
            pTBlk->m_Cmd.m_Port
	);
      }
      else
      {
	TTCP_AddMsg(
	    pmsg,
	    "  Statistics  : UDP <- %s:%d\r\n",
            INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
            htons( pTBlk->m_Cmd.m_RemoteAddr.sin_port )
	);
      }
   }
   else
   {
      if( pTBlk->m_Cmd.m_bTransmit )
      {
	TTCP_AddMsg(
	    pmsg,
	    "  Statistics  : TCP -> %s:%d\r\n",
            INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
            pTBlk->m_Cmd.m_Port
	);
      }
      else
      {
	TTCP_AddMsg(
	    pmsg,
	    "  Statistics  : TCP <- %s:%d\r\n",
            INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
            htons( pTBlk->m_Cmd.m_RemoteAddr.sin_port )
	);
      }
   }

   TTCP_AddMsg(
	pmsg,
	LLU_FMT" bytes in %.2f real seconds = %s/sec +++\r\n",
	pTBlk->m_nbytes,
	realt,
	outfmt(pTBlk->m_nbytes/realt, pTBlk->m_Cmd.m_fmt, buf)
   );

   TTCP_AddMsg(
	pmsg,
	"numCalls: %d; msec/call: %.2f; calls/sec: %.2f\r\n",
	pTBlk->m_numCalls,
	1000.0 * realt/((double )pTBlk->m_numCalls),
	((double )pTBlk->m_numCalls)/realt
   );

   idlet = TTCP_CPUIdleTime( pTBlk );
   if ( idlet >= 0.0 )
   {
       TTCP_AddMsg(
            pmsg,
            "CPU load during transfer: %.2f%%\r\n",
            (100.0 - (idlet*100.0)/realt)
       );
   }
   else
   {
       TTCP_AddMsg(
            pmsg,
            "CPU load during transfer: %s\r\n",
            "not available"
       );
   }
}


// return pointer to TEST_BLOCK associated with a CMD_BLOCK

static PTEST_BLOCK TTCP_AllocateTestBlock( PCMD_BLOCK pCmdBlk )
{
   PTEST_BLOCK pTBlk = (PTEST_BLOCK)pCmdBlk;	// horrible, but true

   pTBlk->m_Socket_fd = INVALID_SOCKET; // fd of network send/receive socket

   return( pTBlk );
}


// free up resources held by a TEST_BLOCK
// (the TEST_BLOCK itself actually gets freed in adi_ttcp_main.c)

static void TTCP_FreeTestBlock( PTEST_BLOCK pTBlk )
{
   if( !pTBlk )
   {
      return;
   }

   if( pTBlk->m_pBufBase )
   {
      free( pTBlk->m_pBufBase );
   }

   pTBlk->m_pBufBase = NULL;

   if( pTBlk->m_Socket_fd != INVALID_SOCKET )
   {
      closesocket( pTBlk->m_Socket_fd );
      pTBlk->m_Socket_fd = INVALID_SOCKET;

   }
}


// error functions

static void TTCP_ExitTestOnNetError( PTEST_BLOCK pTBlk, PCHAR pFnc )
{
   PMSG_BLOCK pmsg = &pTBlk->m_Cmd.m_msgBlk;
   DWORD nError    = TTCP_GetLastError();

   TTCP_AddMsg( pmsg, "*** Network ERROR: %s Failed; ERROR: %d (0x%8.8X)\r\n",
      pFnc, nError, nError );

   TTCP_FreeTestBlock( pTBlk );
   TTCP_TerminateThread( pTBlk );
}


static void TTCP_ExitTestOnCRTError( PTEST_BLOCK pTBlk, PCHAR pFnc )
{
   PMSG_BLOCK pmsg = &pTBlk->m_Cmd.m_msgBlk;

   TTCP_AddMsg( pmsg, "*** CRT ERROR: %s Failed; ERROR: %d (0x%8.8X)\r\n",
      pFnc, errno, errno );

   TTCP_FreeTestBlock( pTBlk );
   TTCP_TerminateThread( pTBlk );
}


// Fill Buffer With Printable Characters...
static void TTCP_FillPattern( PTEST_BLOCK pTBlk )
{
   register char c;
   PCHAR PBPreamble = "PCAUSA PCATTCP Pattern";
   size_t prbm_len = strlen(PBPreamble);
   size_t cnt = (size_t)pTBlk->m_Cmd.m_nBufferSize;
   char  *cp = pTBlk->m_pBuf;

   c = 0;

   //
   // Insert "PCAUSA Pattern" Preamble
   //
   if( cnt > prbm_len )
   {
      strcpy( cp, PBPreamble );
      cnt -= prbm_len;
      cp += prbm_len;
   }

   while( cnt-- > 0 )
   {
      while( !isprint((c&0x7F)) )
      {
         c++;
      }

      *cp++ = (c++&0x7F);
   }
}

static BOOLEAN TTCP_AllocateBuffer( PTEST_BLOCK pTBlk )
{
   PMSG_BLOCK pmsg = &pTBlk->m_Cmd.m_msgBlk;

    //
   // Setup Buffer Configuration
   //
   if( (pTBlk->m_pBufBase =
         (PCHAR)malloc(pTBlk->m_Cmd.m_nBufferSize + pTBlk->m_Cmd.m_nBufAlign))
              == (PCHAR )NULL
     )
   {
      return( FALSE );  // Failed
   }

   //
   // Align The Buffer
   //
   pTBlk->m_pBuf = pTBlk->m_pBufBase;

   if (pTBlk->m_Cmd.m_nBufAlign != 0)
      pTBlk->m_pBuf += (
         pTBlk->m_Cmd.m_nBufAlign
         - ((POINTER_INT)(void*)pTBlk->m_pBuf % pTBlk->m_Cmd.m_nBufAlign
         ) + pTBlk->m_Cmd.m_nBufOffset) % pTBlk->m_Cmd.m_nBufAlign;

   TTCP_AddMsg( pmsg, "  Buffer Size : %d; Alignment: %d/%d\r\n",
      pTBlk->m_Cmd.m_nBufferSize,
      pTBlk->m_Cmd.m_nBufAlign,
      pTBlk->m_Cmd.m_nBufOffset
      );

   return( TRUE );  // Success
}


/////////////////////////////////////////////////////////////////////////////
//// Primary TTCP Functions
//

/////////////////////////////////////////////////////////////////////////////
//// TTCP_TransmitTCP
//
// Purpose
// TTCP TCP transmitter.
//
// Parameters
//   pCmdBlk -         Pointer to CMD_BLOCK that contains options and other
//                     configuration information to be used for the transmit
//                     test.

static DWORD TTCP_TransmitTCP( PCMD_BLOCK pCmdBlk )
{
    PMSG_BLOCK pmsg = &pCmdBlk->m_msgBlk;
    PTEST_BLOCK pTBlk = NULL;
    int   optlen;

    //
    // Say Hello
    //
    TTCP_AddMsg( pmsg, "TCP Transmit Test\r\n" );

    //
    // Allocate Test Instance Data
    // ---------------------------
    // Allocate a TEST_BLOCK for this specific instance. The TEST_BLOCK
    // contains a copy of the calleer's CMD_BLOCK as well as additional
    // members that are used to perform this test.
    //
    pTBlk = TTCP_AllocateTestBlock( pCmdBlk );

    if( !pTBlk )
    {
        TTCP_ExitTestOnCRTError( NULL, "malloc" );

        return( 1 );
    }

    //
    // Setup Local IP Addresses For Test
    //
    pTBlk->m_Cmd.m_LocalAddress.sin_family = AF_INET;
    pTBlk->m_Cmd.m_LocalAddress.sin_port = 0;        /* free choice */

    TTCP_AddMsg( pmsg, "  Transmit    : TCP -> %s:%d\r\n",
        INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
        pTBlk->m_Cmd.m_Port
    );

    //
    // Allocate The Buffer Send/Receive Data
    //
    if( !TTCP_AllocateBuffer( pTBlk ) )
    {
        TTCP_ExitTestOnCRTError( pTBlk, "malloc" );
    }

    //
    // Open Socket For Test
    //
    if( (pTBlk->m_Socket_fd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
    {
        TTCP_ExitTestOnNetError( pTBlk, "socket" );
    }

    //
    // Bind Socket With Local Address
    //
    if( bind(
	    pTBlk->m_Socket_fd,
	    (PSOCKADDR )&pTBlk->m_Cmd.m_LocalAddress,
	    sizeof(pTBlk->m_Cmd.m_LocalAddress)
        ) == SOCKET_ERROR
      )
    {
        TTCP_ExitTestOnNetError( pTBlk, "bind" );
    }

    if( pTBlk->m_Cmd.m_bUseSockOptBufSize )
    {
        if( setsockopt(
		pTBlk->m_Socket_fd,
		SOL_SOCKET,
		SO_SNDBUF,
		(char * )&pTBlk->m_Cmd.m_nSockOptBufSize,
		sizeof pTBlk->m_Cmd.m_nSockOptBufSize
            ) == SOCKET_ERROR
	  )
        {
            TTCP_ExitTestOnNetError( pTBlk, "setsockopt: SO_SNDBUF" );
        }

        TTCP_AddMsg( pmsg,
	             "  SO_SNDBUF   : %d\r\n", pTBlk->m_Cmd.m_nSockOptBufSize );
    }

    //
    // Start TCP Connections
    //

    //
    // We are the client if transmitting
    //
    if( pTBlk->m_Cmd.m_bSockOptDebug )
    {
        if( setsockopt(
		pTBlk->m_Socket_fd,
		SOL_SOCKET,
		SO_DEBUG,
		(PCHAR )&one,     // Boolean
		sizeof(one)
            ) == SOCKET_ERROR
	  )
        {
            TTCP_ExitTestOnNetError( pTBlk, "setsockopt: SO_DEBUG" );
        }
    }

#if SUPPORTS_TCP_NODELAY
    //
    // Set TCP_NODELAY Send Option
    //
    if( pTBlk->m_Cmd.m_nSockOptNoDelay )
    {
        if( setsockopt(
		pTBlk->m_Socket_fd,
		IPPROTO_TCP,
		TCP_NODELAY,
		(PCHAR )&one,     // Boolean
		sizeof(one)
            ) == SOCKET_ERROR
	  )
        {
            DWORD nError = TTCP_GetLastError();
            TTCP_AddMsg( pmsg, "  ERROR: %s: %d (0x%8.8X)\r\n",
          	               "setsockopt: TCP_NODELAY option failed",
	                       nError, nError );
        }
    }

    //
    // Query And Display TCP_NODELAY Send Option
    //
    optlen = sizeof( pTBlk->m_Cmd.m_nSockOptNoDelay );
    if( getsockopt(
	    pTBlk->m_Socket_fd,
	    IPPROTO_TCP,
	    TCP_NODELAY,
	    (PCHAR )&pTBlk->m_Cmd.m_nSockOptNoDelay,
	    &optlen
        ) != SOCKET_ERROR
      )
    {
        if( pTBlk->m_Cmd.m_nSockOptNoDelay )
        {
            TTCP_AddMsg( pmsg, "  TCP_NODELAY : ENABLED (%d)\r\n",
                pTBlk->m_Cmd.m_nSockOptNoDelay );
        }
        else
        {
            TTCP_AddMsg( pmsg, "  TCP_NODELAY : DISABLED (%d)\r\n",
                pTBlk->m_Cmd.m_nSockOptNoDelay );
        }
    }
    else
    {
	DWORD nError = TTCP_GetLastError();
	TTCP_AddMsg( pmsg, "  ERROR: %s: %d (0x%8.8X)\r\n",
			   "getsockopt: TCP_NODELAY option failed",
			   nError, nError );
    }
#endif

    // give receiver a chance to set up
    TTCP_SleepMillisec(TX_WAIT_TIME);

    //
    // Connect To Remote Server
    //
    if( connect(
	    pTBlk->m_Socket_fd,
	    (PSOCKADDR )&pTBlk->m_Cmd.m_RemoteAddr,
	    sizeof(pTBlk->m_Cmd.m_RemoteAddr)
	) == SOCKET_ERROR
      )
    {
        TTCP_ExitTestOnNetError( pTBlk, "connect" );
    }

    TTCP_AddMsg( pmsg, "  Connect     : Connected to %s:%d\r\n",
                       INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
		       pTBlk->m_Cmd.m_Port );

    if( pTBlk->m_Cmd.m_nOptWriteDelay )
    {
        TTCP_AddMsg( pmsg, "  Write Delay : %d microseconds\r\n",
                           pTBlk->m_Cmd.m_nOptWriteDelay);
    }

    if ( pTBlk->m_Cmd.m_period )
    {
	TTCP_AddMsg( pmsg, "  Test Period : %d seconds\r\n",
	                   pTBlk->m_Cmd.m_period / 1000);
    }

    TTCP_FillPattern( pTBlk );

    TTCP_InitStatistics( pTBlk );

    errno = 0;

    if( pTBlk->m_Cmd.m_bOptContinuous )
    {
        TTCP_AddMsg( pmsg, "  Send Mode   : Sending Pattern CONTINUOUS\r\n" );

        if( pTBlk->m_Cmd.m_TokenRate > 0 )
        {
	    // continuous sending at a fixed rate

            int LastTickCount = TTCP_GetMillisecCount();

            TTCP_AddMsg( pmsg, "  Data Rate   : %d bytes/second (%.2f KBps)\r\n",
		               pTBlk->m_Cmd.m_TokenRate,
	                       pTBlk->m_Cmd.m_TokenRate/1000.0 );

	    while ( notExpired(pTBlk) )
	    {
		int ThisTickCount;

		TTCP_SleepMillisec( 1 );
		ThisTickCount = TTCP_GetMillisecCount();
		pTBlk->m_ByteCredits +=
		    ( pTBlk->m_Cmd.m_TokenRate/1000 ) *
		          (ThisTickCount - LastTickCount );

		LastTickCount = ThisTickCount;

		while( (int)pTBlk->m_ByteCredits >= pTBlk->m_Cmd.m_nBufferSize )
		{
		    if( TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize) == pTBlk->m_Cmd.m_nBufferSize )
		    {
			pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;
			pTBlk->m_ByteCredits -= pTBlk->m_Cmd.m_nBufferSize;
		    }
		    else
		    {
			break;
		    }
		}
	    }
        }
        else
        {
	    // continuous sending as fast as possible

            while ( notExpired(pTBlk) &&
		    (TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize)
	                     == pTBlk->m_Cmd.m_nBufferSize) )
                pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;
        }
    }
    else
    {
        TTCP_AddMsg( pmsg,
	             "  Send Mode   : Send Pattern; Number of Buffers: %d\r\n",
                     pTBlk->m_Cmd.m_nNumBuffersToSend);

        if( pTBlk->m_Cmd.m_TokenRate > 0 )
        {
	    // fixed size transmit at fixed rate

            int LastTickCount = TTCP_GetMillisecCount();

            TTCP_AddMsg( pmsg, "  Data Rate   : %d bytes/second (%.2f KBps)\r\n",
 	                       pTBlk->m_Cmd.m_TokenRate,
	                       pTBlk->m_Cmd.m_TokenRate/1000.0 );

            // Pace Transmit Data Rate
            while( notExpired(pTBlk) && pTBlk->m_Cmd.m_nNumBuffersToSend )
            {
                int ThisTickCount;

                TTCP_SleepMillisec( 1 );
                ThisTickCount = TTCP_GetMillisecCount();
                pTBlk->m_ByteCredits +=
		  ( pTBlk->m_Cmd.m_TokenRate/1000 ) * (ThisTickCount - LastTickCount );

                LastTickCount = ThisTickCount;

                while( pTBlk->m_Cmd.m_nNumBuffersToSend > 0 &&
                       (int)pTBlk->m_ByteCredits >= pTBlk->m_Cmd.m_nBufferSize
                     )
                {
                    if( TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize) == pTBlk->m_Cmd.m_nBufferSize )
                    {
                        --pTBlk->m_Cmd.m_nNumBuffersToSend;
                        pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;
                        pTBlk->m_ByteCredits -= pTBlk->m_Cmd.m_nBufferSize;
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }
        else
        {
	    // fixed size transmit as fast as possible

            while ( notExpired(pTBlk) &&
		    pTBlk->m_Cmd.m_nNumBuffersToSend-- &&
		    TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize) == pTBlk->m_Cmd.m_nBufferSize)
                pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;

        }
    }

    if(errno)
        TTCP_ExitTestOnCRTError( pTBlk, "IO" );

    TTCP_LogStatistics( pTBlk );

    TTCP_FreeTestBlock( pTBlk );

    return( 0 );
}



/////////////////////////////////////////////////////////////////////////////
//// TTCP_ReceiveTCP
//
// Purpose
// TTCP TCP stream receiver.
//
// Parameters
//   pCmdBlk - Pointer to CMD_BLOCK that contains options and other
//             configuration information to be used to receive on
//             the stream.
//
// Return Value
// Returns zero(0) for normal return. Non-zero otherwise.
//
// Remarks
// The TTCP_ListenTCP routine listens for connections on the TTCP port.
// When a connection is accepted in TTCP_ListenTCP it then calls this
// function to perform the TCP receive test on the connection socket.
//
// Ownership of the conn_fd socket is passed to this routine, which is
// responsible for eventually closing it.
//

static DWORD TTCP_ReceiveTCP( PCMD_BLOCK pCmdBlk )
{
   PMSG_BLOCK pmsg = &pCmdBlk->m_msgBlk;
   PTEST_BLOCK pTBlk = NULL;
   register int cnt;


   //
   // Allocate Test Instance Data
   // ---------------------------
   // Allocate a TEST_BLOCK for this specific instance. The TEST_BLOCK
   // contains a copy of the calleer's CMD_BLOCK as well as additional
   // members that are used to perform this test.
   //
   pTBlk = TTCP_AllocateTestBlock( pCmdBlk );

   if( !pTBlk )
   {
      closesocket( pCmdBlk->m_ExtraSocket );

      TTCP_ExitTestOnCRTError( NULL, "malloc" );

      return( 1 );
   }

   //
   // Save The Connection Socket
   //
   pTBlk->m_Socket_fd = pCmdBlk->m_ExtraSocket;

   pTBlk->m_Cmd.m_Protocol = IPPROTO_TCP;

   //
   // Allocate The Buffer Send/Receive Data
   //
   if( !TTCP_AllocateBuffer( pTBlk ) )
   {
      TTCP_ExitTestOnCRTError( pTBlk, "malloc" );

      return( 1 );
   }

   TTCP_InitStatistics( pTBlk );
   errno = 0;

   //
   // Discard Received Data
   //
   TTCP_AddMsg( pmsg, "  Receive Mode: Sinking (discarding) Data\r\n" );

   while( (cnt = TTCP_Nread( pTBlk, pTBlk->m_Cmd.m_nBufferSize) ) > 0)
   {
      pTBlk->m_nbytes += cnt;
   }

   if(errno)
      TTCP_ExitTestOnCRTError( pTBlk, "IO" );

   TTCP_LogStatistics( pTBlk );

   TTCP_FreeTestBlock( pTBlk );

   return( 0 );
}


/////////////////////////////////////////////////////////////////////////////
//// TTCP_ListenTCP
//
// Purpose
// TTCP TCP connection listener.
//
// Parameters
//   pCmdBlk - Pointer to CMD_BLOCK that contains options and other
//             configuration information to be used to listen for
//             connections.
//
// Return Value
// Returns zero(0) for normal return. Non-zero otherwise.
//
// Remarks
// The TTCP_ListenTCP routine listens for connections on the TTCP port.
//
// When a connection is accepted TTCP_ListenTCP calls TTCP_ReceiveTCP
// function to perform the actual TCP receive test.
//
// Ownership of the conn_fd socket is passed to TTCP_ReceiveTCP routine,
// which is responsible for eventually closing conn_fd.
//

static DWORD TTCP_ListenTCP( PCMD_BLOCK pCmdBlk )
{
   PMSG_BLOCK pmsg = &pCmdBlk->m_msgBlk;
   PTEST_BLOCK    pTBlk = NULL;
   SOCKET         conn_fd;
   CHAR           szLocalHostName[ 128 ];

   TTCP_AddMsg( pmsg, "TCP Receive Test\r\n" );

   //
   // Allocate Test Instance Data
   // ---------------------------
   // Allocate a TEST_BLOCK for this specific instance. The TEST_BLOCK
   // contains a copy of the calleer's CMD_BLOCK as well as additional
   // members that are used to perform this test.
   //
   pTBlk = TTCP_AllocateTestBlock( pCmdBlk );

   if( !pTBlk )
   {
      TTCP_ExitTestOnCRTError( NULL, "malloc" );

      return( 1 );
   }

   sprintf( szLocalHostName, "Unknown" );
   gethostname( szLocalHostName, 128 );
   TTCP_AddMsg( pmsg, "  Local Host  : %s\r\n", szLocalHostName );

   //
   // Setup Local IP Addresses For Test
   //
   pTBlk->m_Cmd.m_LocalAddress.sin_family = AF_INET;
   pTBlk->m_Cmd.m_LocalAddress.sin_port =  htons( pTBlk->m_Cmd.m_Port );

   //
   // Open Listening Socket
   //
   if( (pTBlk->m_Socket_fd = socket( AF_INET, SOCK_STREAM, 0 ) ) == INVALID_SOCKET )
   {
      TTCP_ExitTestOnNetError( pTBlk, "socket" );
   }

   //
   // Bind Socket With Local Address
   //
   if( bind(
         pTBlk->m_Socket_fd,
         (PSOCKADDR )&pTBlk->m_Cmd.m_LocalAddress,
         sizeof(pTBlk->m_Cmd.m_LocalAddress)
         ) == SOCKET_ERROR
      )
   {
      TTCP_ExitTestOnNetError( pTBlk, "bind" );
   }

   if( pTBlk->m_Cmd.m_bUseSockOptBufSize )
   {
      if( setsockopt(
            pTBlk->m_Socket_fd,
            SOL_SOCKET,
            SO_RCVBUF,
            (char * )&pTBlk->m_Cmd.m_nSockOptBufSize,
            sizeof pTBlk->m_Cmd.m_nSockOptBufSize
            ) == SOCKET_ERROR
         )
      {
         TTCP_ExitTestOnNetError( pTBlk, "setsockopt: SO_RCVBUF" );
      }

      TTCP_AddMsg( pmsg, "  SO_RCVBUF   : %d\r\n", pTBlk->m_Cmd.m_nSockOptBufSize );
   }

   //
   // Start TCP Connections
   // ---------------------
   // We Are The Server.
   //
   listen( pTBlk->m_Socket_fd, 0 );  // allow a queue of 0

   if( pTBlk->m_Cmd.m_bSockOptDebug )
   {
      if( setsockopt(
            pTBlk->m_Socket_fd,
            SOL_SOCKET,
            SO_DEBUG,
            (PCHAR )&one,        // Boolean
            sizeof(one)
            ) == SOCKET_ERROR
         )
      {
         TTCP_ExitTestOnNetError( pTBlk, "setsockopt: SO_DEBUG" );
      }
   }

   pTBlk->m_Cmd.m_nRemoteAddrLen = sizeof( pTBlk->m_Cmd.m_RemoteAddr );

   TTCP_AddMsg( pmsg, "  Listening...: On port %d\r\n", pTBlk->m_Cmd.m_Port );

   if( ( conn_fd = accept( pTBlk->m_Socket_fd, (PSOCKADDR )&pTBlk->m_Cmd.m_RemoteAddr,
         &pTBlk->m_Cmd.m_nRemoteAddrLen) ) == SOCKET_ERROR)
   {
      TTCP_ExitTestOnNetError( pTBlk, "accept" );
   }
   else
   {
      SOCKADDR_IN peer;
      int peerlen = sizeof(peer);

      if (getpeername( conn_fd, (PSOCKADDR ) &peer,
            &peerlen) == SOCKET_ERROR)
      {
         TTCP_ExitTestOnNetError( pTBlk, "getpeername" );
      }

      pCmdBlk->m_RemoteAddr.sin_addr.s_addr = peer.sin_addr.s_addr;
      pCmdBlk->m_RemoteAddr.sin_port = peer.sin_port;

      TTCP_AddMsg( pmsg, "  Accept      : TCP <- %s:%d\r\n",
            INET_NTOA( pCmdBlk->m_RemoteAddr.sin_addr ),
            htons( pCmdBlk->m_RemoteAddr.sin_port ) );
   }

   // close the listen socket
   closesocket(pTBlk->m_Socket_fd);

   pTBlk->m_Cmd.m_ExtraSocket = conn_fd;

   TTCP_ReceiveTCP( &pTBlk->m_Cmd );
   TTCP_FreeTestBlock( pTBlk );

   return( 0 );
}


/////////////////////////////////////////////////////////////////////////////
//// TTCP_TransmitUDP
//
// Purpose
// TTCP UDP transmitter.
//
// Parameters
//   pCmdBlk -         Pointer to CMD_BLOCK that contains options and other
//                     configuration information to be used for the transmit
//                     test.
//
// Return Value
// Returns zero(0) for normal return. Non-zero otherwise.
//

static DWORD TTCP_TransmitUDP( PCMD_BLOCK pCmdBlk )
{
    PMSG_BLOCK pmsg = &pCmdBlk->m_msgBlk;
    PTEST_BLOCK pTBlk = NULL;

    //
    // Say Hello
    //
    TTCP_AddMsg( pmsg, "UDP Transmit Test\r\n" );

    //
    // Allocate Test Instance Data
    // ---------------------------
    // Allocate a TEST_BLOCK for this specific instance. The TEST_BLOCK
    // contains a copy of the calleer's CMD_BLOCK as well as additional
    // members that are used to perform this test.
    //
    pTBlk = TTCP_AllocateTestBlock( pCmdBlk );

    if( !pTBlk )
    {
        TTCP_ExitTestOnCRTError( NULL, "malloc" );

        return( 1 );
    }

    //
    // Setup Local IP Addresses For Test
    //
    pTBlk->m_Cmd.m_LocalAddress.sin_family = AF_INET;
    pTBlk->m_Cmd.m_LocalAddress.sin_port = 0;        /* free choice */

    TTCP_AddMsg( pmsg, "  Transmit    : UDP -> %s:%d\r\n",
        INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
        pTBlk->m_Cmd.m_Port );

    //
    // Setup Buffer Configuration
    //
    if( pTBlk->m_Cmd.m_nBufferSize <= UDP_GUARD_BUFFER_LENGTH )
    {
        pTBlk->m_Cmd.m_nBufferSize = UDP_GUARD_BUFFER_LENGTH + 1; // send more than the sentinel size
    }

    //
    // Allocate The Buffer Send/Receive Data
    //
    if( !TTCP_AllocateBuffer( pTBlk ) )
    {
        TTCP_ExitTestOnCRTError( pTBlk, "malloc" );
    }

    //
    // Open Socket For Test
    //
    if( (pTBlk->m_Socket_fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == INVALID_SOCKET
        )
    {
        TTCP_ExitTestOnNetError( pTBlk, "socket" );
    }

    //
    // Bind Socket With Local Address
    //
    if( bind(
        pTBlk->m_Socket_fd,
        (PSOCKADDR )&pTBlk->m_Cmd.m_LocalAddress,
        sizeof(pTBlk->m_Cmd.m_LocalAddress)
        ) == SOCKET_ERROR
        )
    {
        TTCP_ExitTestOnNetError( pTBlk, "bind" );
    }

    if( pTBlk->m_Cmd.m_bUseSockOptBufSize )
    {
        if( setsockopt(
            pTBlk->m_Socket_fd,
            SOL_SOCKET,
            SO_SNDBUF,
            (char * )&pTBlk->m_Cmd.m_nSockOptBufSize,
            sizeof pTBlk->m_Cmd.m_nSockOptBufSize
            ) == SOCKET_ERROR
            )
        {
            TTCP_ExitTestOnNetError( pTBlk, "setsockopt: SO_SNDBUF" );
        }

        TTCP_AddMsg( pmsg, "  SO_SNDBUF   : %d\r\n", pTBlk->m_Cmd.m_nSockOptBufSize );
    }

    if( pTBlk->m_Cmd.m_nOptWriteDelay )
    {
        TTCP_AddMsg( pmsg, "  Write Delay : %d microseconds\r\n",
            pTBlk->m_Cmd.m_nOptWriteDelay
            );
    }

    if ( pTBlk->m_Cmd.m_period )
    {
	TTCP_AddMsg( pmsg, "  Test Period : %d seconds\r\n",
	                   pTBlk->m_Cmd.m_period / 1000);
    }

    TTCP_FillPattern( pTBlk );

    // give receiver a chance to set up
    TTCP_SleepMillisec(TX_WAIT_TIME);

    TTCP_InitStatistics( pTBlk );

    errno = 0;

    TTCP_Nwrite( pTBlk, UDP_GUARD_BUFFER_LENGTH ); /* rcvr start */

    if( pTBlk->m_Cmd.m_bOptContinuous )
    {
        int LastTickCount = TTCP_GetMillisecCount();

        TTCP_AddMsg( pmsg, "  Send Mode   : Send Pattern CONTINUOUS\r\n" );

        if( pTBlk->m_Cmd.m_TokenRate > 0 )
        {
	    // continuous transmit at fixed rate

            TTCP_AddMsg(
		pmsg,
		"  Data Rate   : %d bytes/second (%.2f KBps)\r\n",
		pTBlk->m_Cmd.m_TokenRate, pTBlk->m_Cmd.m_TokenRate/1000.0
	    );

            while( notExpired(pTBlk) )
            {
                int ThisTickCount;

                TTCP_SleepMillisec( 1 );
                ThisTickCount = TTCP_GetMillisecCount();
                pTBlk->m_ByteCredits +=
		    ( pTBlk->m_Cmd.m_TokenRate/1000 ) *
			(ThisTickCount - LastTickCount );

                LastTickCount = ThisTickCount;

                while( (int)pTBlk->m_ByteCredits >= pTBlk->m_Cmd.m_nBufferSize )
                {
                    if( TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize) == pTBlk->m_Cmd.m_nBufferSize )
                    {
                        pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;
                        pTBlk->m_ByteCredits -= pTBlk->m_Cmd.m_nBufferSize;
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }
        else
        {
	    // continuous transmit as fast as possible

            while ( notExpired(pTBlk) &&
		    TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize) == pTBlk->m_Cmd.m_nBufferSize)
                pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;
        }
    }
    else
    {
        TTCP_AddMsg(
	    pmsg,
	    "  Send Mode   : Send Pattern; Number of Buffers: %d\r\n",
            pTBlk->m_Cmd.m_nNumBuffersToSend
	);

        if( pTBlk->m_Cmd.m_TokenRate > 0 )
        {
	    // fixed size transmit at fixed rate

            int LastTickCount = TTCP_GetMillisecCount();

            TTCP_AddMsg(
		pmsg,
		"  Data Rate   : %d bytes/second (%.2f KBps)\r\n",
		pTBlk->m_Cmd.m_TokenRate, pTBlk->m_Cmd.m_TokenRate/1000.0
	    );

            // Pace Transmit Data Rate
            while( notExpired(pTBlk) && pTBlk->m_Cmd.m_nNumBuffersToSend )
            {
                int ThisTickCount;

                TTCP_SleepMillisec( 1 );
                ThisTickCount = TTCP_GetMillisecCount();
                pTBlk->m_ByteCredits +=
		    ( pTBlk->m_Cmd.m_TokenRate/1000 ) *
			(ThisTickCount - LastTickCount );

                LastTickCount = ThisTickCount;

                while( pTBlk->m_Cmd.m_nNumBuffersToSend > 0
                    && (int)pTBlk->m_ByteCredits >= pTBlk->m_Cmd.m_nBufferSize
                    )
                {
                    if( TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize) == pTBlk->m_Cmd.m_nBufferSize )
                    {
                        --pTBlk->m_Cmd.m_nNumBuffersToSend;
                        pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;
                        pTBlk->m_ByteCredits -= pTBlk->m_Cmd.m_nBufferSize;
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }
        else
        {
	    // fixed size transmit as fast as possible

            while ( notExpired(pTBlk) &&
		    pTBlk->m_Cmd.m_nNumBuffersToSend-- &&
		    TTCP_Nwrite( pTBlk, pTBlk->m_Cmd.m_nBufferSize) == pTBlk->m_Cmd.m_nBufferSize)
                pTBlk->m_nbytes += pTBlk->m_Cmd.m_nBufferSize;
        }
    }

    TTCP_Nwrite( pTBlk, UDP_GUARD_BUFFER_LENGTH ); /* rcvr end */

    if(errno)
        TTCP_ExitTestOnCRTError( pTBlk, "IO" );

    TTCP_LogStatistics( pTBlk );

    //
    // Pause To Allow Receiver To Flush Receive Buffers
    // ------------------------------------------------
    // PCATTCP Version 2.01.01.06 and prior (as well as the original TTCP
    // application)were subject to failure at various points. This was due
    // to the fact that UDP has no flow control, so a UDP transfer overwhelms
    // the receiving computer. The reason it was failing was that the end marker
    // of the data transfer was getting lost in the network while the receiving
    // computer was being overwhelmed with data. This would cause the receiving
    // computer to hang, waiting for more data, but the transmitting computer
    // had already finished sending all the data along with the end markers.
    //
    // The folks at Clarkson University identified a simple fix for this
    // weakness which PCAUSA has adopted in V2.01.01.07. See the URL:
    //
    // http://www.clarkson.edu/projects/itl/HOWTOS/mcnair_summer2002/Paper/User_Manual_PCATTCP_Controller_1.00.htm
    //
    // A simple pause by the transmitter before the end markers were sent would
    // allow the receiving computer time to free its internal receiving buffers
    // and obtain an end marker.
    //
    TTCP_SleepMillisec( 1000 );

    TTCP_Nwrite( pTBlk, UDP_GUARD_BUFFER_LENGTH );   // rcvr end
    TTCP_Nwrite( pTBlk, UDP_GUARD_BUFFER_LENGTH );   // rcvr end
    TTCP_Nwrite( pTBlk, UDP_GUARD_BUFFER_LENGTH );   // rcvr end
    TTCP_Nwrite( pTBlk, UDP_GUARD_BUFFER_LENGTH );   // rcvr end

    TTCP_FreeTestBlock( pTBlk );

    return( 0 );
}


/////////////////////////////////////////////////////////////////////////////
//// TTCP_ReceiveUDP
//
// Purpose
// TTCP UDP datagram receiver.
//
// Parameters
//   pCmdBlk - Pointer to CMD_BLOCK that contains options and other
//             configuration information.
//
// Return Value
// Returns zero(0) for normal return. Non-zero otherwise.
//
// Remarks
// The TTCP_ReceiveUDP routine performs the TTCP UDP receive test.
//

static DWORD TTCP_ReceiveUDP( PCMD_BLOCK pCmdBlk )
{
   PMSG_BLOCK pmsg = &pCmdBlk->m_msgBlk;
   PTEST_BLOCK pTBlk = NULL;
   BOOLEAN     bContinue = TRUE;

   TTCP_AddMsg( pmsg, "UDP Receive Test\r\n" );

   //
   // Allocate Test Instance Data
   // ---------------------------
   // Allocate a TEST_BLOCK for this specific instance. The TEST_BLOCK
   // contains a copy of the calleer's CMD_BLOCK as well as additional
   // members that are used to perform this test.
   //
   pTBlk = TTCP_AllocateTestBlock( pCmdBlk );

   if( !pTBlk )
   {
      TTCP_ExitTestOnCRTError( NULL, "malloc" );

      return( 1 );
   }

   TTCP_AddMsg( pmsg, "  Protocol    : UDP\r\n" );
   TTCP_AddMsg( pmsg, "  Port        : %d\r\n", pTBlk->m_Cmd.m_Port );

   //
   // Setup Local IP Addresses For Test
   //
   pTBlk->m_Cmd.m_LocalAddress.sin_family = AF_INET;
   pTBlk->m_Cmd.m_LocalAddress.sin_port =  htons( pTBlk->m_Cmd.m_Port );

   //
   // Setup Buffer Configuration
   //
   if( pTBlk->m_Cmd.m_nBufferSize <= UDP_GUARD_BUFFER_LENGTH
      )
   {
      pTBlk->m_Cmd.m_nBufferSize = UDP_GUARD_BUFFER_LENGTH + 1; // send more than the sentinel size
   }

   //
   // Allocate The Buffer Send/Receive Data
   //
   if( !TTCP_AllocateBuffer( pTBlk ) )
   {
      TTCP_ExitTestOnCRTError( pTBlk, "malloc" );
   }

   //
   // Open Socket For Test
   //
   if( (pTBlk->m_Socket_fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == INVALID_SOCKET )
   {
      TTCP_ExitTestOnNetError( pTBlk, "socket" );
   }

   //
   // Bind Socket With Local Address
   //
   if( bind(
         pTBlk->m_Socket_fd,
         (PSOCKADDR )&pTBlk->m_Cmd.m_LocalAddress,
         sizeof(pTBlk->m_Cmd.m_LocalAddress)
         ) == SOCKET_ERROR
      )
   {
      TTCP_ExitTestOnNetError( pTBlk, "bind" );
   }

   if( pTBlk->m_Cmd.m_bUseSockOptBufSize )
   {
      if( setsockopt(
            pTBlk->m_Socket_fd,
            SOL_SOCKET,
            SO_RCVBUF,
            (char * )&pTBlk->m_Cmd.m_nSockOptBufSize,
            sizeof pTBlk->m_Cmd.m_nSockOptBufSize
            ) == SOCKET_ERROR
         )
      {
         TTCP_ExitTestOnNetError( pTBlk, "setsockopt: SO_RCVBUF" );
      }

      TTCP_AddMsg( pmsg, "  SO_RCVBUF   : %d\r\n", pTBlk->m_Cmd.m_nSockOptBufSize );
   }

   while( bContinue )
   {
      register int cnt;
      int going = 0;
      BOOLEAN bIsFirst = TRUE;

      TTCP_InitStatistics( pTBlk );
      errno = 0;

      memset(
         &pTBlk->m_Cmd.m_RemoteAddr,
         0x00,
         sizeof( IN_ADDR )
         );

      while( (cnt=TTCP_Nread( pTBlk, pTBlk->m_Cmd.m_nBufferSize)) > 0 )
      {
         if( cnt <= UDP_GUARD_BUFFER_LENGTH )
         {
            if( going )
            {
               going = 0;
               break;    /* "EOF" */
            }

            going = 1;
            TTCP_InitStatistics( pTBlk );
         }
         else
         {
            if( bIsFirst )
            {
               TTCP_AddMsg(
		    pmsg,
		    "  recvfrom    : UDP <- %s:%d\r\n",
		    INET_NTOA( pTBlk->m_Cmd.m_RemoteAddr.sin_addr ),
		    htons( pTBlk->m_Cmd.m_RemoteAddr.sin_port )
		);

               bIsFirst = FALSE;
            }

            pTBlk->m_nbytes += cnt;
         }
      }

      if(errno)
         TTCP_ExitTestOnCRTError( pTBlk, "IO" );

      if( pTBlk->m_nbytes > 0 )
      {
         TTCP_LogStatistics( pTBlk );
      }

      bContinue = FALSE;
   }

   TTCP_FreeTestBlock( pTBlk );

   return( 0 );
}




////////////////////////////////////////////////////////////////////////////////
// TRANSFER thread entrypoint                                                 //
////////////////////////////////////////////////////////////////////////////////
int TTCP_DispatchTransfer(void* arg)
{
    PCMD_BLOCK  pcmd = (PCMD_BLOCK)arg;
    PTEST_BLOCK ptfr = (PTEST_BLOCK)arg;

    //
    // Dispatch To Specialized Procedures
    //
    if( pcmd->m_bTransmit )
    {
        DWORD nResult;

        TTCP_SleepMillisec(2000);  // allow receive end time to set up

        if( pcmd->m_Protocol == IPPROTO_UDP )
        {
            nResult = TTCP_TransmitUDP( pcmd );
        }
        else
        {
            nResult = TTCP_TransmitTCP( pcmd );
        }
    }
    else
    {
        DWORD nResult;

	if( pcmd->m_Protocol == IPPROTO_UDP )
	{
	    nResult = TTCP_ReceiveUDP( pcmd );
	}
	else
	{
	    nResult = TTCP_ListenTCP( pcmd );
	}
    }

    TTCP_TerminateThread(ptfr);

    return (0);
}
